/* Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.filter;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.util.Hashtable;

import org.cresques.geo.ViewPortData;
import org.cresques.io.data.RasterBuf;
import org.cresques.io.datastruct.Statistic;
import org.cresques.px.Extent;


/**
 * Filtro para raster. Ancestro de todos los filtros.
 * @author Luis W. Sevilla (sevilla_lui@gva.es)
 */
public abstract class RasterFilter implements IRasterFilter {
	/**
	 * Nombre del filtro. Este nombre debe asignarlo cada clase de filtro especifico y es necesario
	 * para leer el nombre en el listado de fitros de la pila.
	 */
	protected String		filterName = "";
    protected RasterBuf 	raster = null;
    protected Statistic 	stats = null;
    protected Image 		image = null;
    protected int 			height = 0;
    protected int 			width = 0;
    protected Hashtable 	params = new Hashtable();
    protected boolean 		isSupersampling = false;
    protected ViewPortData	viewPortData = null;
    protected Extent		extent = null;
    /**
     * Variable que control la aplicacin o no del filtro. Si est a false aunque est en 
     * la pila el filtro no se ejecutar.
     */
    protected boolean 		exec = true; 
    
    /**
     * Vectores con los pasos usados para el supersampling
     */
    protected int[]			stepX = null;
    protected int[]			stepY = null;
    /**
     * Posicin de los vectores de pasos que se est analizando en un instante dado
     */
    protected int			contX = -1, contY = -1;

    /**
     * @author Nacho Brodin (brodin_ign@gva.es)
     */
    public static class Kernel{
    	public double[][] kernel = null;
    	protected double divisor = 0;
    	
    	public Kernel(double[][] k){
    		this.kernel = k;
    		
    		for (int i=0;i< kernel.length;i++)
    			for (int j=0;j<kernel[0].length;j++)
    				divisor = divisor + kernel[i][j];
    	}
    	
    	public Kernel(double[][] k,double divisor){
    		this.kernel = k;
    		this.divisor = divisor;
    	}
    	
    	public double kernelOperation(Kernel k){
    		double res = 0;
    		for (int i=0;i< kernel.length;i++)
    			for (int j=0;j<kernel[0].length;j++)
    				res += kernel[i][j] *  k.kernel[i][j];
    		return res;
    	}
    	
    	public double convolution (Kernel k){
    		double res =0;
    		res = this.kernelOperation(k);
    		if (this.divisor !=0)
    			res = res/divisor;
    		return Math.abs(res);
    	}

		public double getDivisor() {
			return divisor;
		}

		public void setDivisor(double divisor) {
			this.divisor = divisor;
		}
		
		public int getLado(){
			return kernel.length;
		}
    }


    /**
     * Constructor
     */
    public RasterFilter() {
    }

    /**
     * Aplica el filtro sobre el raster pasado pixel a pixel
     */
    public void execute() {
        pre();
        if (exec) {
        	if(isSupersampling()){
        		contX = -1;
        		for (int col = 0; col < width; col += stepX[contX]){
        			contY = -1;
        			for (int line = 0; line < height; line += stepY[contY]){ 
        				processSuperSampling(col, line);	
        				contY ++;
        			}
        			contX ++;
        		}
        	}else{
        		for (int col = 0; col < width; col += 1)
        			for (int line = 0; line < height; line += 1) 
        				process(col, line);
        	}
        }
        post();
    }

    /**
     * Aplica el filtro sobre el raster pasado por lineas
     */
    public void executeLines() {
        if (exec) {
            for (int y = 0; y < height; y++)
                processLine(y);
        }
    }

    /**
     * Aade un parmetro al filtro
     * @param name	Clave del parmetro
     * @param param Objeto pasado como parmetro
     */
    public void addParam(String name, Object param) {
        params.put(name, param);
    }

    /**
     * Elimina un parmetro del filtro
     * @param name Clave del parmetro a eliminar
     */
    public void removeParam(String name){
    	params.remove(name);
    }
    
    /**
     * Obtiene un parmetro a partir de la clave
     * @param name Parmetro
     * @return Parmetro
     */
    public Object getParam(String name){
    	return params.get(name);
    }
    
    /**
     * Funcin que contiene el cdigo a ejecutar antes de aplicar el filtro
     */
    abstract public void pre();

    /**
     * Funcin que contiene el cdigo a ejecutar despues de aplicar el filtro
     */
    abstract public void post();

    /**
     * Ejecucin del filtro para un pixel de la imagen
     */
    abstract public void process(int x, int y);
    
    /**
     * Ejecucin del filtro para un pixel de la imagen
     */
    abstract public void processSuperSampling(int x, int y);

    /**
     * Ejecucin del filtro para una lnea de la imagen
     */
    abstract public void processLine(int y);

    /**
     * Obtiene el tipo de datos del raster de entrada
     */
    abstract public int getInRasterDataType();

    /**
     * Obtiene el tipo de datos del raster de salida
     */
    abstract public int getOutRasterDataType();

    /**
     * Obtiene el resultado del filtro despues de su ejecucin a travs de una clave
     * @param name        clave para obtener un objeto resultado del filtro.
     */
    abstract public Object getResult(String name);

	/**
	 * @param viewPortData The viewPortData to set.
	 */
	public void setViewPortData(ViewPortData viewPortData) {
		this.viewPortData = viewPortData;
	}
	/**
	 * @param extent The extent to set.
	 */
	public void setExtent(Extent extent) {
		this.extent = extent;
	}

	/**
	 * Obtiene true si el filtro va a ser ejecutado o false si no va a serlo
	 * @return
	 */
	public boolean isExec() {
		return exec;
	}

	/**
	 * Asigna el valor a la variable exec. Esta estar a true si el filtro se ejecutar la prxima
	 * vez que se repinte o false si no se ejecuta.
	 * @param exec
	 */
	public void setExec(boolean exec) {
		this.exec = exec;
	}
	
	/**
	 * Asigna el valor del paso (Solo cuando se hace supersampling).
	 * @param stepx paso en X
	 * @param stepy paso en Y
	 */
	public void setStep(int[] stepX, int[] stepY){
		if(stepX != null && stepY != null){
			isSupersampling = true;
			this.stepX = stepX;
			this.stepY = stepY;
		}else{
			isSupersampling = false;
			stepX = stepY = null;
		}
	}
	
	/**
	 * Obtiene si la imagen pasada tiene supersampling o no
	 * @return true si en el dibujado se ha aplicado supersampling y false si no se ha aplicado.
	 */
	protected boolean isSupersampling(){
		return isSupersampling;
	}

	/**
	 * Obtiene el nombre del filtro. Este nombre debe asignarlo cada clase 
	 * de filtro especifico y es necesario para leer el nombre en el listado de fitros de la pila.
	 * @return Cadena que representa el nombre del filtro. 
	 */
	public String getFilterName() {
		return filterName;
	}
	
	/**
	 * ASigna el nombre del filtro. 
	 * @param fName Cadena que representa el identificador del filtro 
	 */
	public void setFilterName(String fName) {
		filterName = fName;
	}
}