/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 * 
 * Copyright (C) 2004-5. 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 * 
 * cresques@gmail.com
 */
package org.cresques.io;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.DataBuffer;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.TreeMap;

import org.cresques.cts.ICoordTrans;
import org.cresques.cts.IProjection;
import org.cresques.filter.PixelFilter;
import org.cresques.filter.SimplePixelFilter;
import org.cresques.io.data.BandList;
import org.cresques.io.data.RasterBuf;
import org.cresques.io.data.RasterMetaFileTags;
import org.cresques.io.datastruct.Metadata;
import org.cresques.io.datastruct.Palette;
import org.cresques.io.exceptions.SupersamplingNotSupportedException;
import org.cresques.px.Extent;
import org.cresques.px.IObjList;
import org.cresques.px.PxContour;
import org.cresques.px.PxObjList;
import org.gvsig.i18n.Messages;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

/**
 * Manejador de ficheros raster georeferenciados.
 * 
 * Esta clase abstracta es el ancestro de todas las clases que proporcionan
 * soporte para ficheros raster georeferenciados.<br>
 * Actua tambien como una 'Fabrica', ocultando al cliente la manera en que
 * se ha implementado ese manejo. Una clase nueva que soportara un nuevo
 * tipo de raster tendra que registrar su extensin o extensiones usando
 * el mtodo @see registerExtension.<br> 
 * @author "Luis W. Sevilla" <sevilla_lui@gva.es>*
 */

public abstract class GeoRasterFile extends GeoFile {
	
	/**
	 * Flag que representa a la banda del Rojo
	 */
	public static final int 	RED_BAND	= 0x01;
	
	/**
	 * Flag que representa a la banda del Verde
	 */
	public static final int 	GREEN_BAND	= 0x02;
	
	/**
	 * Flag que representa a la banda del Azul
	 */
	public static final int 	BLUE_BAND	= 0x04;
	public static TreeMap 		supportedExtensions = null;
	protected Component 		updatable = null;
	protected boolean 			doTransparency = false;
	private boolean				verifySize = false;
	/**
	 * Parmetros de transformacin del fichero .rmf. Esta ser distinta
	 * de la identidad si la funcin rmfExists() devuelve true.
	 */
	protected AffineTransform	rmfTransform = new AffineTransform();
	
	/**
	 * Esta variable estar a true si el driver est supersampleando en el ltimo dibujado.
	 * Debe tenerse especial cuidado ya que est consulta no es adecuada desde gvSIG ya que el
	 * dibujado se realiza asincronamente por lo que el valor puede no coincidir con el ltimo
	 * dibujado.
	 */
	protected boolean			isSupersampling = false;
	/**
	 * Paso correspondiente al supersampling o subsampling que se est aplicando en el ltimo dibujado.
	 */
	protected int[] stepArrayX = null, stepArrayY = null;
	
	/**
	 * Filtro para raster.
	 * Permite eliminar la franja inutil alrededor de un raster girado o de
	 * un mosaico de borde irregular.
	 * 
	 * Funciona bien solo con raster en tonos de gris, porque se basa que
	 * el valor del pixel no supere un determinado valor 'umbral' que se
	 * le pasa al constructor.
	 * 
	 * Desarrollado para 'limpiar' los bordes de los mosaicos del SIG
	 * Oleicola. Para ese caso los parmetros del constructo son:
	 * PixelFilter(0x10ffff00, 0xff000000, 0xf0f0f0);
	 */
	protected PixelFilter 		tFilter = null;
	
	/**
	 * Asignacin de banda del Rojo a una banda de la imagen
	 */
	protected int 				rBandNr = 1;
	
	/**
	 * Asignacin de banda del Verde a una banda de la imagen
	 */
	protected int 				gBandNr = 1;
	
	/**
	 * Asignacin de banda del Azul a una banda de la imagen
	 */
	protected int 				bBandNr = 1;
	
	/**
	 * Nmero de bandas de la imagen
	 */
	protected int 				bandCount = 1;
	private int 				dataType = DataBuffer.TYPE_BYTE;
	protected Palette			palette = null;
	/**
	 * Parmetros de transformacin del fichero .rmf. Estas variables tendrn valores distinto
	 * de 0 si la funcin rmfExists() devuelve true.
	 */
	protected double imageWidth = 0D, imageHeight = 0D;
	
	static {
		Messages.addResourceFamily("org.cresques.translations.text", "org.cresques.ui");
		supportedExtensions = new TreeMap();
		String os = System.getProperty("os.name");
		supportedExtensions.put("ecw",  EcwFile.class);
		supportedExtensions.put("jp2",  EcwFile.class);
		
		supportedExtensions.put("sid",  MrSidFile.class);

		supportedExtensions.put("bmp", GdalFile.class);
		supportedExtensions.put("gif", GdalFile.class);
		supportedExtensions.put("img", GdalFile.class);
		supportedExtensions.put("tif", GdalFile.class);
		supportedExtensions.put("tiff", GdalFile.class);
		supportedExtensions.put("jpg", GdalFile.class);
		supportedExtensions.put("png", GdalFile.class);
		supportedExtensions.put("vrt", GdalFile.class);
		
		supportedExtensions.put("dat",  GdalFile.class); // Envi
		supportedExtensions.put("lan",  GdalFile.class); // Erdas
		supportedExtensions.put("gis",  GdalFile.class); // Erdas
		supportedExtensions.put("pix",  GdalFile.class); // PCI Geomatics
		supportedExtensions.put("aux",  GdalFile.class); // PCI Geomatics
		//if(!os.toLowerCase().startsWith("windows"))
		supportedExtensions.put("adf",  GdalFile.class); // ESRI Grids
		supportedExtensions.put("mpr",  GdalFile.class); // Ilwis
		supportedExtensions.put("mpl",  GdalFile.class); // Ilwis
		supportedExtensions.put("map",  GdalFile.class); // PC Raster
		supportedExtensions.put("asc",  GdalFile.class); // PC Raster
		supportedExtensions.put("vrt",  GdalFile.class); 
	}
	
	/**
	 * Factoria para abrir distintos tipos de raster.
	 * 
	 * @param proj Proyeccin en la que est el raster.
	 * @param fName Nombre del fichero.
	 * @return GeoRasterFile, o null si hay problemas.
	 */
	public static GeoRasterFile openFile(IProjection proj, String fName) {
		String ext = fName.toLowerCase().substring(fName.lastIndexOf('.')+1);
		GeoRasterFile grf = null;

		//if (!supportedExtensions.containsKey(ext)) 
			//return grf;
		Class clase = null;
		if (supportedExtensions.containsKey(ext))
			clase = (Class) supportedExtensions.get(ext);
		else
			clase = GdalFile.class;
			
		
		Class [] args = {IProjection.class, String.class};
		try {
			Constructor hazNuevo = clase.getConstructor(args);
			Object [] args2 = {proj, fName};
			grf = (GeoRasterFile) hazNuevo.newInstance(args2);
			grf.setFileSize(new File(fName).length());
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			System.err.println("Extension not supported!!!");
			return null;
		}
		return grf;
	}
	
	/**
	 * Registra una clase que soporta una extensin raster.
	 * @param ext extensin soportada.
	 * @param clase clase que la soporta.
	 */
	public static void registerExtension(String ext, Class clase) {
		ext = ext.toLowerCase();
		System.out.println("RASTER: extension '"+ext+"' supported.");
		supportedExtensions.put(ext, clase);
	}
	
	/**
	 * Tipo de fichero soportado.
	 * Devuelve true si el tipo de fichero (extension) est soportado, si no
	 * devuelve false.
	 * 
	 * @param fName Fichero raster
	 * @return  true si est soportado, si no false.
 	 */
	public static boolean fileIsSupported(String fName) {
		//return true;
		String ext = fName.toLowerCase().substring(fName.lastIndexOf('.') + 1);
		return supportedExtensions.containsKey(ext);
	}
	
	/**
	 * Constructor
	 * @param proj	Proyeccin
	 * @param name	Nombre del fichero de imagen.
	 */
	public GeoRasterFile(IProjection proj, String name) {
		super(proj, name);
	}
	
	/**
	 * Carga un fichero raster. Puede usarse para calcular el extent e instanciar 
	 * un objeto de este tipo.
	 */
	abstract public GeoFile load();
	
	/**
	 * Cierra el fichero y libera los recursos.
	 */
	abstract public void close();
	
	/**
	 * Obtiene la codificacin del fichero XML
	 * @param file Nombre del fichero XML
	 * @return Codificacin
	 */
	private String readFileEncoding(String file){
		FileReader fr;
		String encoding = null;
		try
    	{
			fr = new FileReader(file);
    		BufferedReader br = new BufferedReader(fr);
    		char[] buffer = new char[100];
    		br.read(buffer);
    		StringBuffer st = new StringBuffer(new String(buffer));
    		String searchText = "encoding=\"";
    		int index = st.indexOf(searchText);
    		if (index>-1) {
    			st.delete(0, index+searchText.length());
    			encoding = st.substring(0, st.indexOf("\""));
    		}
    		fr.close();
    	} catch(FileNotFoundException ex)	{
    		ex.printStackTrace();
    	} catch (IOException e) {
			e.printStackTrace();
		}
    	return encoding;
	}
	
	private double[] parserExtent(KXmlParser parser) throws XmlPullParserException, IOException {		
		double originX = 0D, originY = 0D, w = 0D, h = 0D;
		double pixelSizeX = 0D, pixelSizeY = 0D;
		double shearX = 0D, shearY = 0D;
		
		boolean end = false;
    	int tag = parser.next();
    	while (!end) {
    		switch(tag) {
        		case KXmlParser.START_TAG:
        			if(parser.getName() != null){	
						if (parser.getName().equals(RasterMetaFileTags.POSX)){
							originX = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.POSY)){
							originY = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.PX_SIZE_X)){
							pixelSizeX = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.PX_SIZE_Y)){
							pixelSizeY = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.ROTX)){
							shearX = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.ROTY)){
							shearY = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.WIDTH)){
							w = Double.parseDouble(parser.nextText());
						}else if (parser.getName().equals(RasterMetaFileTags.HEIGHT)){
							h = Double.parseDouble(parser.nextText());
						}
					}		    	    
					break;
         		case KXmlParser.END_TAG:
         			if (parser.getName().equals(RasterMetaFileTags.BBOX))
         				end = true;
        			break;
        		case KXmlParser.TEXT:
        			break;
    		}
    		tag = parser.next();
    	}
		
    	double[] values = {originX, originY, w, h, pixelSizeX, pixelSizeY, shearX, shearY};
		return values;
	}
	
	/**
	 * Obtiene la informacin de georreferenciacin asociada a la imagen en un fichero .rmf. Esta 
	 * georreferenciacin tiene la caracteristica de que tiene prioridad sobre la de la imagen.
	 * Es almacenada en la clase GeoFile en la variable virtualExtent.
	 * @param file Fichero de metadatos .rmf
	 */
	protected void readGeoInfo(String file){
		String rmf = file.substring(0, file.lastIndexOf(".") + 1) + "rmf";
		File rmfFile = new File(rmf);
		if(!rmfFile.exists())
			return;
		
		boolean georefOk = false;
		
		FileReader fr = null;
		String v = null;
		try {
			fr = new FileReader(rmf);
			KXmlParser parser = new KXmlParser();
			parser.setInput(new FileInputStream(rmf), readFileEncoding(rmf));
			int tag = parser.nextTag();
			if ( parser.getEventType() != KXmlParser.END_DOCUMENT ){    		
				parser.require(KXmlParser.START_TAG, null, RasterMetaFileTags.MAIN_TAG);    			
				while(tag != KXmlParser.END_DOCUMENT) {
					switch(tag) {
						case KXmlParser.START_TAG:
							if (parser.getName().equals(RasterMetaFileTags.LAYER)) {
								int layerListTag = parser.next();
								boolean geoRefEnd = false;
								while (!geoRefEnd){
									if(parser.getName() != null){
										if (parser.getName().equals(RasterMetaFileTags.PROJ)){
											//System.out.println("PROJ:"+parser.nextText());
										} else if (parser.getName().equals(RasterMetaFileTags.BBOX)){
											double[] values = parserExtent(parser);
											rmfTransform = new AffineTransform(	values[4], values[7],
													                           	values[6], values[5],
													   							values[0], values[1]);
											georefOk = true;
										} else if (parser.getName().equals(RasterMetaFileTags.DIM)){
											boolean DimEnd = false;
											while (!DimEnd){
												layerListTag = parser.next();
												if(parser.getName() != null){	
													if (parser.getName().equals(RasterMetaFileTags.PX_WIDTH)){
														imageWidth = Double.parseDouble(parser.nextText());
													}else if (parser.getName().equals(RasterMetaFileTags.PX_HEIGHT)){
														imageHeight = Double.parseDouble(parser.nextText());
														DimEnd = true;
													}													
												}
											}
											geoRefEnd = true;
										}
									}
									layerListTag = parser.next();
								}
							}
							break;
						case KXmlParser.END_TAG:							
							break;
						case KXmlParser.TEXT:							
							break;
					}
					tag = parser.next();
				}
				parser.require(KXmlParser.END_DOCUMENT, null, null);
			}
			
			if(georefOk){
				rmfExists = true;
				setExtentTransform(	rmfTransform.getTranslateX(), rmfTransform.getTranslateY(), 
									rmfTransform.getScaleX(), rmfTransform.getScaleY());
				createExtentsFromRMF(	rmfTransform.getTranslateX(), rmfTransform.getTranslateY(), 
										rmfTransform.getScaleX(), rmfTransform.getScaleY(), 
										imageWidth, imageHeight, 
										rmfTransform.getShearX(), rmfTransform.getShearY());
			}
			
		} catch (FileNotFoundException fnfEx) {
		} catch (XmlPullParserException xmlEx) {
			xmlEx.printStackTrace();
		} catch (IOException e) {
		} 
		try{
			if(fr != null)
				fr.close();
		}catch(IOException ioEx){
			//No est abierto el fichero por lo que no hacemos nada
		}
	}

	/**
	 * Asigna una transformacin al raster para que se tenga en cuenta en la asignacin del setView. 
	 * Esta asignacin recalcula el extent, el requestExtent y asigna el AffineTransform que se 
	 * usar para la transformacin. Esta transformacin ser considerada como si la imagen tuviera 
	 * asociado un rmf.
	 * @param t Transformacin afn a aplicar
	 */
	public void setAffineTransform(AffineTransform t){
		rmfExists = true;
		rmfTransform = (AffineTransform)t.clone();
		setExtentTransform(t.getTranslateX(), t.getTranslateY(), t.getScaleX(), t.getScaleY());
		createExtentsFromRMF(	t.getTranslateX(), t.getTranslateY(), t.getScaleX(), t.getScaleY(), 
								this.getWidth(), this.getHeight(), 
								t.getShearX(), t.getShearY());
	}
	
	/**
	 * Asigna una transformacin al raster para que se tenga en cuenta en la asignacin del setView. 
	 * Esta asignacin recalcula el extent, el requestExtent y asigna el AffineTransform que se 
	 * usar para la transformacin. Esta transformacin ser considerada como si la imagen tuviera 
	 * asociado un rmf.
	 * @param originX Coordenada X de origen del raster
	 * @param originY Coordenada Y de origen del raster
	 * @param pixelSizeX Tamao de pixel en X
	 * @param pixelSizeY Tamao de pixel en Y
	 * @param imageWidth Ancho del raster en pixels
	 * @param imageHeight Alto del raster en pixels
	 * @param shearX Shearing en X
	 * @param shearY Shearing en Y
	 */
	public void setAffineTransform(	double originX, double originY, double pixelSizeX, 
										double pixelSizeY, double shearX, double shearY){
		rmfExists = true;
		rmfTransform.setToTranslation(originX, originY);
		rmfTransform.shear(shearX, shearY);
		rmfTransform.scale(pixelSizeX, pixelSizeY);
		setExtentTransform(originX, originY, pixelSizeX, pixelSizeY);
		createExtentsFromRMF(	originX, originY, pixelSizeX, pixelSizeY, 
								imageWidth, imageHeight, shearX, shearY);
	}
	
	/**
	 * Obtiene la matriz de transformacin que se aplica sobre la visualizacin 
	 * del raster.
	 * @return Matriz de transformacin.
	 */
	public AffineTransform getAffineTransform(){
		return rmfTransform;
	}
	
	/**
	 * Elimina la matriz de transformacin asociada al raster y que se tiene en cuenta para
	 * el setView. Este reseteo tendr en cuenta que si el raster tiene asociado un rmf
	 * esta transformacin no ser eliminada sino que se asignar la correspondiente al rmf
	 * existente.  
	 * @return devuelve true si tiene fichero rmf asociado y false si no lo tiene.
	 */
	public boolean resetAffineTransform(){
		rmfExists = false;
		rmfTransform.setToIdentity();
		
		//Crea los extent iniciales
		load();
		
		//Lee y carga el rmf si existe
		readGeoInfo(this.getName());
		
		if(rmfExists)
			return true;
		else
			return false;
	}
	
	/**
	 * <P>
	 * Calcula el extent de la imagen a partir del fichero rmf con y sin rotacin. El extent con rotacin corresponde
	 * a la variable extent que contiene el extent verdadero marcado por el fichero de georreferenciacin .rmf. El extent
	 * sin rotacin requestExtent es utilizado para realizar la peticin ya que la peticin al driver no se puede
	 * hacer con coordenadas rotadas.
	 * 
	 * El calculo de la bounding box rotada lo hace con los valores de transformacin leidos desde el fichero .rmf.
	 * </p>
	 * <P>
	 * Para el calculo de una esquina aplicamos la formula siguiente:<BR>
	 * PtoX = originX + pixelSizeX * x + shearX * y;<BR>
	 * PtoY = originY + shearY * x + pixelSizeY * y;<BR>
	 * Aplicandolo a las cuatro esquinas sustituimos en cada una de ellas por.
	 * </P>
	 * <UL> 
	 * <LI>Esquina superior izquierda: x = 0; y = 0;</LI>
	 * <LI>Esquina superior derecha: x = MaxX; y = 0;</LI>
	 * <LI>Esquina inferior izquierda: x = 0; y = MaxY;</LI>
	 * <LI>Esquina inferior derecha: x = MaxX; y = MaxY;</LI>
	 * </UL> 
	 * <P>
	 * quedandonos en los cuatro casos:
	 * </P>
	 * <UL> 
	 * <LI>Esquina superior izquierda: originX; originY;</LI>
	 * <LI>Esquina superior derecha: PtoX = originX + pixelSizeX * x; PtoY = originY + shearY * x;</LI>
	 * <LI>Esquina inferior izquierda:  PtoX = originX + shearX * y; PtoY = originY + pixelSizeY * y;</LI>
	 * <LI>Esquina inferior derecha: PtoX = originX + pixelSizeX * x + shearX * y; PtoY = originY + shearY * x + pixelSizeY * y;</LI>
	 * </UL>
	 * 
	 * <P>
	 * El calculo de la bounding box se realizar de la misma forma pero anulando los parametros de shearing.
	 * </P>
	 * 
	 * @param originX Coordenada X de origen del raster
	 * @param originY Coordenada Y de origen del raster
	 * @param pixelSizeX Tamao de pixel en X
	 * @param pixelSizeY Tamao de pixel en Y
	 * @param imageWidth Ancho del raster en pixels
	 * @param imageHeight Alto del raster en pixels
	 * @param shearX Shearing en X
	 * @param shearY Shearing en Y
	 */
	private void createExtentsFromRMF(	double originX, double originY, double pixelSizeX, double pixelSizeY, 
										double imageWidth, double imageHeight, double shearX, double shearY){
				
		Point2D p1 = new Point2D.Double(originX, originY);
		Point2D p2 = new Point2D.Double(originX + shearX * imageHeight, originY + pixelSizeY * imageHeight);
		Point2D p3 = new Point2D.Double(originX + pixelSizeX * imageWidth, originY + shearY * imageWidth);
		Point2D p4 = new Point2D.Double(originX + pixelSizeX * imageWidth + shearX * imageHeight, originY + pixelSizeY * imageHeight + shearY * imageWidth);
		
		double minX = Math.min(Math.min(p1.getX(), p2.getX()), Math.min(p3.getX(), p4.getX()));
		double minY = Math.min(Math.min(p1.getY(), p2.getY()), Math.min(p3.getY(), p4.getY()));
		double maxX = Math.max(Math.max(p1.getX(), p2.getX()), Math.max(p3.getX(), p4.getX()));
		double maxY = Math.max(Math.max(p1.getY(), p2.getY()), Math.max(p3.getY(), p4.getY()));
		extent = new Extent(minX, minY, maxX, maxY);
		requestExtent = new Extent(originX, originY, originX + (pixelSizeX * imageWidth), originY + (pixelSizeY * imageHeight));
	}
	
	/**
	 * Calcula la transformacin que se produce sobre la vista cuando la imagen tiene un fichero .rmf
	 * asociado. Esta transformacin tiene diferencias entre los distintos formatos por lo que debe calcularla
	 * el driver correspondiente.
	 * @param originX Origen de la imagen en la coordenada X
	 * @param originY Origen de la imagen en la coordenada Y
	 */
	abstract public void setExtentTransform(double originX, double originY, double psX, double psY);
	
	public static PxContour getContour(String fName, String name, IProjection proj) {
		PxContour contour = null;
		return contour;
	}
		
	/**
	 * Obtiene el ancho de la imagen
	 * @return Ancho de la imagen
	 */
	abstract public int getWidth();
	
	/**
	 * Obtiene el ancho de la imagen
	 * @return Ancho de la imagen
	 */
	abstract public int getHeight();

	/**
	 * Reproyeccin.
	 * @param rp	Coordenadas de la transformacin
	 */
	abstract public void reProject(ICoordTrans rp);

	/**
	 * Asigna un nuevo Extent 
	 * @param e	Extent
	 */
	abstract public void setView(Extent e);
	
	/**
	 * Obtiene el extent asignado
	 * @return	Extent
	 */
	abstract public Extent getView();
	
	public void setTransparency(boolean t) {
		doTransparency = t;
		tFilter = new PixelFilter(255);
	}
	
	/**
	 * Asigna un valor de transparencia
	 * @param t	Valor de transparencia
	 */
	public void setTransparency(int t ) {
		doTransparency = true;
		tFilter = new SimplePixelFilter(255 - t);
	}
	
	public boolean getTransparency() { return doTransparency; }
	
	public void setAlpha(int alpha) {
		if (!doTransparency) setTransparency(255 - alpha);
		else tFilter.setAlpha(alpha);
	}
	public int getAlpha() {
		if (tFilter == null)
			return 255;
		return tFilter.getAlpha();
	}
	
	public void setUpdatable(Component c) { updatable = c; }
	
	/**
	 * Actualiza la imagen
	 * @param width	ancho
	 * @param height	alto
	 * @param rp	Reproyeccin
	 * @return	img
	 */
	abstract public Image updateImage(int width, int height, ICoordTrans rp);

	/**
	 * Obtiene el valor del raster en la coordenada que se le pasa.
	 * El valor ser Double, Int, Byte, etc. dependiendo del tipo de
	 * raster.
	 * @param x	coordenada X
	 * @param y coordenada Y
	 * @return
	 */
	abstract public Object getData(int x, int y, int band);

	/**
	 * Actualiza la/s banda/s especificadas en la imagen.
	 * @param width		ancho
	 * @param height	alto
	 * @param rp		reproyeccin
	 * @param img		imagen
	 * @param flags		que bandas [ RED_BAND | GREEN_BAND | BLUE_BAND ]
	 * @return		img
	 * @throws SupersamplingNotSupportedException
	 */
	abstract public Image updateImage(int width, int height, ICoordTrans rp, Image img, int origBand, int destBand)throws SupersamplingNotSupportedException;

	public int getBandCount() { return bandCount; }
	
	/**
	 * Asocia un colorBand al rojo, verde o azul.
	 * @param flag cual (o cuales) de las bandas.
	 * @param nBand	que colorBand
	 */
	
	public void setBand(int flag, int bandNr) {
		if ((flag & GeoRasterFile.RED_BAND) == GeoRasterFile.RED_BAND) rBandNr = bandNr;
		if ((flag & GeoRasterFile.GREEN_BAND) == GeoRasterFile.GREEN_BAND) gBandNr = bandNr;
		if ((flag & GeoRasterFile.BLUE_BAND) == GeoRasterFile.BLUE_BAND) bBandNr = bandNr;
	}

	/**
	 * Devuelve el colorBand activo en la banda especificada.
	 * @param flag banda.
	 */
	
	public int getBand(int flag) {
		if (flag == GeoRasterFile.RED_BAND) return rBandNr;
		if (flag == GeoRasterFile.GREEN_BAND) return gBandNr;
		if (flag == GeoRasterFile.BLUE_BAND) return bBandNr;
		return -1;
	}
	
	/**
	 * @return Returns the dataType.
	 */
	public int getDataType() {
		return dataType;
	}
	
	/**
	 * @param dataType The dataType to set.
	 */
	public void setDataType(int dataType) {
		this.dataType = dataType;
	}

	public IObjList getObjects() {
		// TODO hay que aadir el raster a la lista de objetos
		IObjList oList = new PxObjList(proj);
		return oList;
	}
	
	/**
	 * Calcula los parmetros de un worl file a partir de las esquinas del raster.
	 *    1. X pixel size A
	 *    2. X rotation term D
	 *    3. Y rotation term B
	 *    4. Y pixel size E
	 *    5. X coordinate of upper left corner C
	 *    6. Y coordinate of upper left corner F
	 * where the real-world coordinates x',y' can be calculated from
	 * the image coordinates x,y with the equations
	 *  x' = Ax + By + C and y' = Dx + Ey + F.
	 *  The signs of the first 4 parameters depend on the orientation
	 *  of the image. In the usual case where north is more or less
	 *  at the top of the image, the X pixel size will be positive
	 *  and the Y pixel size will be negative. For a south-up image,
	 *  these signs would be reversed.
	 * 
	 * You can calculate the World file parameters yourself based
	 * on the corner coordinates. The X and Y pixel sizes can be
	 *  determined simply by dividing the distance between two
	 *  adjacent corners by the number of columns or rows in the image.
	 *  The rotation terms are calculated with these equations:
	 * 
	 *  # B = (A * number_of_columns + C - lower_right_x') / number_of_rows * -1
	 *  # D = (E * number_of_rows + F - lower_right_y') / number_of_columns * -1
	 * 
	 * @param corner (tl, tr, br, bl)
	 * @return
	 */
	public static double [] cornersToWorldFile(Point2D [] esq, Dimension size) {
		double a=0,b=0,c=0,d=0,e=0,f=0;
		double x1 = esq[0].getX(), y1 = esq[0].getY();
		double x2 = esq[1].getX(), y2 = esq[1].getY();
		double x3 = esq[2].getX(), y3 = esq[2].getY();
		double x4 = esq[3].getX(), y4 = esq[3].getY();
		// A: X-scale
		a = Math.abs( Math.sqrt( (x1-x2)*(x1-x2)+(y1-y2)*(y1-y2))
		      / size.getWidth());

		// E: negative Y-scale
		e =  - Math.abs(Math.sqrt((x1-x4)*(x1-x4)+
		      (y1-y4)*(y1-y4))/size.getHeight());

		// C, F: upper-left coordinates
		c = x1;
		f = y1;
		
		// B & D: rotation parameters
		b = (a * size.getWidth() + c - x3 ) / size.getHeight() * -1;
		d = (e * size.getHeight() + f - y3 ) / size.getWidth() * -1;

		double [] wf = {a,d,b,e,c,f}; 
		return wf;  
	}
    public static String printWF(String fName, Point2D [] esq, Dimension sz) {
    	double [] wf = GeoRasterFile.cornersToWorldFile(esq, sz);
    	System.out.println("wf para "+fName);
    	System.out.println(esq+"\n"+sz);
    	String wfData = "";
    	for (int i=0; i<6; i++)
    		wfData += wf[i]+"\n";
		System.out.println(wfData);
		return wfData;
    }
    
    public static void saveWF(String fName, String data) throws IOException {
    	FileWriter fw = new FileWriter(fName);
    	fw.write(data);
    	fw.flush();
    	fw.close();
    }

	/**
	 * Cosulta si hay que verificar la relacin de aspecto de la imagen, es decir comprueba que el ancho/alto
	 * pasados a updateImage coinciden con el ancho/alto solicitado en setView a la imagen
	 * @return true si est verificando la relacin de aspecto. 
	 */
	public boolean mustVerifySize() {
		return verifySize;
	}

	/**
	 * Asigna el flag que dice si hay que verificar la relacin de aspecto de la imagen, es decir 
	 * comprueba que el ancho/alto pasados a updateImage coinciden con el ancho/alto solicitado 
	 * en setView a la imagen.
	 * @return true si est verificando la relacin de aspecto. 
	 */
	public void setMustVerifySize(boolean verifySize) {
		this.verifySize = verifySize;
	}

	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * No aplica supersampleo ni subsampleo sino que devuelve una matriz de igual tamao a los
	 * pixeles de disco. 
	 * @param x Posicin X superior izquierda
	 * @param y Posicin Y superior izquierda
	 * @param w Ancho en coordenadas reales
	 * @param h Alto en coordenadas reales
	 * @param rasterBuf	Buffer de datos
	 * @param bandList
	 * @return Buffer de datos
	 */
	abstract public RasterBuf getWindowRaster(double x, double y, double w, double h, BandList bandList, RasterBuf rasterBuf);
	
	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas pixel. 
	 * No aplica supersampleo ni subsampleo sino que devuelve una matriz de igual tamao a los
	 * pixeles de disco. 
	 * @param x Posicin X superior izquierda
	 * @param y Posicin Y superior izquierda
	 * @param w Ancho en coordenadas reales
	 * @param h Alto en coordenadas reales
	 * @param rasterBuf	Buffer de datos
	 * @param bandList
	 * @return Buffer de datos
	 */
	abstract public RasterBuf getWindowRaster(int x, int y, int w, int h, BandList bandList, RasterBuf rasterBuf);

	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * Se aplica supersampleo o subsampleo dependiendo del tamao del buffer especificado.
	 * 
	 * @param minX Posicin mnima X superior izquierda
	 * @param minY Posicin mnima Y superior izquierda
	 * @param maxX Posicin mxima X inferior derecha
	 * @param maxY Posicin mxima Y inferior derecha
	 * @param bufWidth Ancho del buffer de datos
	 * @param bufHeight Alto del buffer de datos
	 * @param rasterBuf	Buffer de datos
	 * @param bandList
	 * @return Buffer de datos
	 */
	abstract public RasterBuf getWindowRaster(double minX, double minY, double maxX, double maxY, int bufWidth, int bufHeight, BandList bandList, RasterBuf rasterBuf);
	
	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * No aplica supersampleo ni subsampleo sino que devuelve una matriz de igual tamao a los
	 * pixeles de disco. 
	 * @param x Posicin X superior izquierda
	 * @param y Posicin Y superior izquierda
	 * @param w Ancho en coordenadas reales
	 * @param h Alto en coordenadas reales
	 * @param rasterBuf	Buffer de datos
	 * @param bandList
	 * @return Buffer de datos
	 */
	abstract public RasterBuf getWindowRasterWithNoData(double x, double y, double w, double h, BandList bandList, RasterBuf rasterBuf);
	
	abstract public byte[] getWindow(int ulX, int ulY, int sizeX, int sizeY, int band);
	
	abstract public int getBlockSize();
	
	/**
	 * Obtiene el objeto que contiene los metadatos. Este mtodo debe ser redefinido por los
	 * drivers si necesitan devolver metadatos. 
	 * @return
	 */
	public Metadata getMetadata(){
		return null;
	}
	
	/**
	 * Asigna un extent temporal que puede coincidir con el de la vista. Esto es 
	 * util para cargar imagenes sin georreferenciar ya que podemos asignar el extent
	 * que queramos para ajustarnos a una vista concreta
	 * @param tempExtent The tempExtent to set.
	 */
	public void setExtent(Extent ext) {
		this.extent = ext;
	}
	
	/**
	 * Dice si el fichero tiene georreferenciacin o no.
	 * @return true si tiene georreferenciacin y false si no la tiene
	 */
	public boolean isGeoreferenced(){
		return true;
	}

	/**
	 * Informa de si el driver ha supersampleado en el ltimo dibujado. Es el driver el que colocar
	 * el valor de esta variable cada vez que dibuja. Debe tenerse especial cuidado ya que est consulta no es adecuada desde gvSIG ya que el
	 * dibujado se realiza asincronamente por lo que el valor puede no coincidir con el ltimo
	 * dibujado.
	 * @return true si se ha supersampleado y false si no se ha hecho.
	 */
	public boolean isSupersampling() {
		return this.isSupersampling;
	}
	
	/**
	 * Asigna el valor del paso en X e Y aplicado en el ltimo dibujado. Es el driver el que colocar
	 * el valor de esta variable cada vez que dibuja. 
	 * @param stepx Paso en X
	 * @param stepy Paso en Y
	 */
	public void setStep(int[] stepArrayx, int[] stepArrayy){
		this.stepArrayX = stepArrayx;
		this.stepArrayY = stepArrayy;
	}
	
	/**
	 * Obtiene el valor del paso en X aplicado en el ltimo dibujado. Es el driver el que colocar
	 * el valor de esta variable cada vez que dibuja. 
	 * @return
	 */
	public int[] getStepX(){
		return stepArrayX;
	}
	
	/**
	 * Obtiene el valor del paso en Y aplicado en el ltimo dibujado. Es el driver el que colocar
	 * el valor de esta variable cada vez que dibuja. 
	 * @return
	 */
	public int[] getStepY(){
		return stepArrayY;
	}

	/**
	 * Obtiene el objeto paleta. Esta paleta es la que tiene adjunta el fichero de disco. Si es
	 * null este objeto quiere decir que no tiene paleta para su visualizacin. 
	 * @return Palette
	 */
	public Palette getPalette() {
		return palette;
	}

	/**
	 * Asigna el objeto paleta. Esta paleta es la que tiene adjunta el fichero de disco. Si es
	 * null este objeto quiere decir que no tiene paleta para su visualizacin. 
	 * @param Palette
	 */
	public void setPalette(Palette palette) {
		this.palette = palette;
	}
	
	/**
	 * Mtodo que indica si existe un fichero .rmf asociado al GeoRasterFile.
	 * @return
	 */
	public boolean rmfExists(){
		return this.rmfExists;
	}
		
	/**
	 * Obtiene los parmetros de la transformacin afn que corresponde con los elementos de
	 * un fichero tfw.
	 * <UL> 
	 * <LI>[1]tamao de pixel en X</LI>
	 * <LI>[2]rotacin en X</LI>
	 * <LI>[4]rotacin en Y</LI>
	 * <LI>[5]tamao de pixel en Y</LI>
	 * <LI>[0]origen en X</LI>
	 * <LI>[3]origen en Y</LI>
	 * </UL>
	 * Este mtodo debe ser reimplementado por el driver si tiene esta informacin. En principio
	 * Gdal es capaz de proporcionarla de esta forma.
	 * @return vector de double con los elementos de la transformacin afn.
	 */
	public double[] getTransform(){return null;}
	
	/**
	 * Convierte un punto desde coordenadas pixel a coordenadas del mundo.
	 * @param pt Punto a transformar
	 * @return punto transformado en coordenadas del mundo
	 */
	abstract public Point2D rasterToWorld(Point2D pt);
	
	/**
	 * Convierte un punto desde del mundo a coordenadas pixel.
	 * @param pt Punto a transformar
	 * @return punto transformado en coordenadas pixel
	 */
	abstract public Point2D worldToRaster(Point2D pt);
	
	/**
	 * Vuelve a leer la paleta del fichero por si ha sido modificada
	 *
	 */
	public void readPalette(){}
}
