/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 */
package org.cresques.io.data;

import java.awt.geom.Point2D;
import java.io.File;
import java.util.ArrayList;

import org.cresques.io.GeoRasterFile;
import org.cresques.io.exceptions.BandFoundInListException;
import org.cresques.io.exceptions.FileFoundInListException;
import org.cresques.io.exceptions.NotSupportedExtensionException;
import org.cresques.px.Extent;

/**
 * Clase que representa una imagen de raster georreferenciada formada por varias
 * imagenes de disco que tienen la misma extensin. Contiene funcionalidades para 
 * abrir ficheros, gestionar el extent, pintar el raster sobre un DataImage con 
 * su gestin de bandas correspondiente.
 *  
 * @author Nacho Brodin (brodin_ign@gva.es)
 *
 */
public class GeoRasterMultiFile{
	//File list
	private ArrayList 		files = new ArrayList();
	private String			name = null;
	//Band list
	private BandList		bandList = new BandList();
			
	public GeoRasterMultiFile(String name){
		this.name = name;
	}
	
	/**
	 * Add a file to the list.
	 * @param f file to add.
	 */
	public void addFile(GeoRasterFile f)throws FileFoundInListException{
		if(findFile(f))
			throw new FileFoundInListException("The file already is in list.");
		files.add(f);
		addBands(f);
	}
	
	/**
	 * Aade un fichero a la lista a partir de su nombre
	 * @param f fichero a aadir.
	 */
	public void addFile(String fileName)throws FileFoundInListException, NotSupportedExtensionException{
		if(findFile(fileName))
			throw new FileFoundInListException("The file already is in list.");
		GeoRasterFile f = GeoRasterFile.openFile(null, fileName);
		files.add(f);
		addBands(f);
	}
	
	/**
	 * Aade el fichero a lista de georrasterfiles y sus bandas a la lista de bandas
	 * @param grf
	 */
	private void addBands(GeoRasterFile grf){
		if(grf == null)
			return;
		
		int dataType = grf.getDataType();
		for(int i = 0; i < grf.getBandCount();i++){
			try{
				Band band = new Band(grf.getName(), i, dataType);
				bandList.addBand(band, i);
			}catch(BandFoundInListException ex){
				//No aadimos la banda
			}
		}
	}
	
	/**
	 * Elimina un fichero a la lista a partir de su nombre
	 * @param fileName	Nombre del fichero a eliminar.
	 */
	public void removeFile(String fileName){
		for(int i=0;i<files.size();i++){
			if(((GeoRasterFile)files.get(i)).getName().equals(fileName)){
				files.remove(i);
				bandList.removeBands(fileName);
				return;
			}
		}
	}
	
	/**
	 * Elimina un fichero a la lista
	 * @param file Fichero a eliminar
	 */
	public void removeFile(GeoRasterFile file){
		for(int i=0;i<files.size();i++){
			if(((GeoRasterFile)files.get(i)).getName().equals(file.getName())){
				files.remove(i);
				bandList.removeBands(file.getName());
				return;
			}
		}
	}
		
	/**
	 * Obtiene el nmero de ficheros en la lista
	 * @return integer.
	 */
	public int getFileCount(){
		return files.size();
	}
	
	/**
	 * Encuentra un fichero en la lista.
	 * @param file Fichero bscado.
	 * @return true si se ha hallado el fichero y false si no se 
	 * ha encontrado
	 */
	public boolean findFile(GeoRasterFile file){
		for(int i = 0;i<files.size();i++){
			GeoRasterFile grf = (GeoRasterFile)files.get(i); 
			if(	grf.getName().equals(file.getName()))
				return true;
		}
		return false;
	}
	
	/**
	 * Encuentra un fichero en la lista.
	 * @param file Fichero bscado.
	 * @return true si se ha hallado el fichero y false si no se 
	 * ha encontrado
	 */
	public boolean findFile(String fileName){
		for(int i = 0;i<files.size();i++){
			if(((GeoRasterFile)files.get(i)).getName().equals(fileName))
				return true;
		}
		return false;
	}
	    		
	/**
	 * @see org.javaGeoRaster.io.GeoData
	 */
	public void close(){
		for(int i = 0; i < files.size(); i++)
			((GeoRasterFile)files.get(i)).close();
	}
	
	/**
	 * Obtiene en un array de String la lista de nombres de ficheros
	 * @return lista de nombres de los ficheros del GeoRasterMultiFile
	 */
	public String[] getNameFileStringList(){
		String[] list = new String[files.size()];
		for(int i = 0; i < files.size(); i++)
			list[i] = ((GeoRasterFile)files.get(i)).getName();
		return list;
	}
	
	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * No aplica supersampleo ni subsampleo sino que devuelve una matriz de igual tamao a los
	 * pixeles de disco. 
	 * @param x Posicin X superior izquierda
	 * @param y Posicin Y superior izquierda
	 * @param w Ancho en coordenadas reales
	 * @param h Alto en coordenadas reales
	 * @return Buffer de datos
	 */
	public RasterBuf getWindowRaster(double x, double y, double w, double h) {
		
		Extent selectedExtent = new Extent(x, y, x + w, y - h);

		int width = (int)((selectedExtent.width() * ((GeoRasterFile)files.get(0)).getWidth()) / ((GeoRasterFile)files.get(0)).getExtent().width());
		int height = (int)((selectedExtent.height() * ((GeoRasterFile)files.get(0)).getHeight()) / ((GeoRasterFile)files.get(0)).getExtent().height());
					
		RasterBuf raster = new RasterBuf(getDataType()[0], width, height, bandList.getDrawableBandsCount(), false);
		
		for(int i = 0; i < files.size(); i++)
			raster = ((GeoRasterFile)files.get(i)).getWindowRaster(x, y, w, h, bandList, raster);
				
		return raster;
	}
	
	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * Aplica supersampleo o subsampleo en funcin del tamao del buffer. Esta operacin la gestiona
	 * el driver.
	 * @param minX Valor mnimo de la X en coordenadas reales
	 * @param minY Valor mnimo de la Y en coordenadas reales
	 * @param maxX Valor mximo de la X en coordenadas reales
	 * @param maxY Valor mximo de la Y en coordenadas reales
	 * @param bufWidth ancho del buffer lde datos
	 * @param bufHeight alto del buffer de datos
	 * @param bandList
	 * @return Buffer de datos
	 */
	public RasterBuf getWindowRaster(double minX, double minY, double maxX, double maxY, 
									int bufWidth, int bufHeight) {
		RasterBuf raster = new RasterBuf(getDataType()[0], bufWidth, bufHeight, bandList.getDrawableBandsCount(), true);
		
		for(int i = 0; i < files.size(); i++)
			raster = ((GeoRasterFile)files.get(i)).getWindowRaster(minX, minY, maxX, maxY, bufWidth, bufHeight, bandList, raster);
							
		return raster;
	}
	
	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * No aplica supersampleo ni subsampleo sino que devuelve una matriz de igual tamao a los
	 * pixeles de disco. 
	 * @param x Posicin X superior izquierda
	 * @param y Posicin Y superior izquierda
	 * @param w Ancho en coordenadas pixel
	 * @param h Alto en coordenadas pixel
	 * @param bandList
	 * @return Buffer de datos
	 */
	public RasterBuf getWindowRaster(int x, int y, int w, int h) {
				
		RasterBuf raster = new RasterBuf(getDataType()[0], w, h, bandList.getDrawableBandsCount(), true);
		
		for(int i = 0; i < files.size(); i++)
			raster = ((GeoRasterFile)files.get(i)).getWindowRaster(x, y, w, h, bandList, raster);
							
		return raster;
	}
	
	/**
	 * Obtiene una ventana de datos de la imagen a partir de coordenadas reales. 
	 * No aplica supersampleo ni subsampleo sino que devuelve una matriz de igual tamao a los
	 * pixeles de disco. 
	 * @param x Posicin X superior izquierda
	 * @param y Posicin Y superior izquierda
	 * @param w Ancho en coordenadas reales
	 * @param h Alto en coordenadas reales
	 * @return Buffer de datos
	 */
	public RasterBuf getWindowRasterWithNoData(double x, double y, double w, double h) {
		
		Extent selectedExtent = new Extent(x, y, x + w, y - h);

		//Leemos pixels completos aunque el valor obtenido sea decimal. Esto se consigue redondeando
		//por arriba el ms alto y por abajo el menor y luego restandolos
		
		Point2D p1 = ((GeoRasterFile)files.get(0)).worldToRaster(new Point2D.Double(x, y));
		Point2D p2 = ((GeoRasterFile)files.get(0)).worldToRaster(new Point2D.Double(x + w, y - h));
		int width = (int)Math.abs(Math.ceil(p2.getX()) - Math.floor(p1.getX())); 
		int height = (int)Math.abs(Math.floor(p1.getY()) - Math.ceil(p2.getY()));
		
		RasterBuf raster = new RasterBuf(getDataType()[0], width, height, bandList.getDrawableBandsCount(), true);
		
		//Si no vamos a ajustar el extent al raster inicializamos el buffer a noData ya que este puede ser
		//ms grande y salirse de los lmites.
		initBufferToNoData(raster, bandList);			
				
		for(int i = 0; i < files.size(); i++)
			raster = ((GeoRasterFile)files.get(i)).getWindowRasterWithNoData(x, y, w, h, bandList, raster);
							
		return raster;
	}
			
	/**
	 * Inicializa el buffer a valores NoData
	 * @param raster Buffer a inicializar
	 * @param bandList Lista de bandas
	 */
	private void initBufferToNoData(RasterBuf raster, BandList bandList){
		for(int i = 0; i < bandList.getDrawableBandsCount(); i++){
			switch(getDataType()[0]){
			case RasterBuf.TYPE_BYTE:raster.assign(i, raster.getByteNoDataValue());break;
			case RasterBuf.TYPE_SHORT:raster.assign(i, raster.getShortNoDataValue());break;
			case RasterBuf.TYPE_INT:raster.assign(i, raster.getIntNoDataValue());break;
			case RasterBuf.TYPE_FLOAT:raster.assign(i, raster.getFloatNoDataValue());break;
			case RasterBuf.TYPE_DOUBLE:raster.assign(i, raster.getNoDataValue());break;
			}
		}	
	}
	//******************************
	//Setters and Getters
	//******************************
	
	/**
	 * Calcula el tamao de los ficheros en disco
	 * @return tamao en bytes de todos los ficheros de la lista
	 */
	public long getFileSize(){
		int len = 0;
		for(int i=0;i<files.size();i++){
			if(((GeoRasterFile)files.get(i)) != null){
				File f = new File(((GeoRasterFile)files.get(i)).getName());
				len += f.length();
			}
		}
		return len;
	}
	
	/**
	 * Obtiene la altura de la imagen a partir de la primera
	 * @return altura
	 */
	public int getHeight() {
		for(int i=0;i<files.size();i++)
			if(((GeoRasterFile)files.get(i)) != null)
				return ((GeoRasterFile)files.get(i)).getHeight();
		return 0;
	}

	/**
	 * Obtiene la anchura de la imagen a partir de la primera
	 * @return anchura
	 */
	public int getWidth() {
		for(int i=0;i<files.size();i++)
			if(((GeoRasterFile)files.get(i)) != null)
				return ((GeoRasterFile)files.get(i)).getWidth();
		return 0;	}
	
	/**
	 * Obtiene el nmero de bandas del fichero
	 * @return
	 */
	public int getBandCount(){
		return bandList.getBandCount();
	}

	/**
	 * Obtiene el tipo de dato por banda
	 * @return tipo de dato por banda
	 */
	public int[] getDataType() {
		int[] dt = new int[getFileCount()];
		for(int i=0;i<files.size();i++)
			dt[i] = ((GeoRasterFile)files.get(i)).getDataType();
				
	    if(dt.length == 0)
	    	return null;
	    else
	    	return dt;
	}
	
	/**
	 * Obtiene fichero de la posicin i.
	 * @param i Posicin del fichero a obtener.
	 * @return GeoRasterFileDataset.
	 */
	public GeoRasterFile getFile(int i){
		return (GeoRasterFile)files.get(i);
	}
	
	/**
	 * Obtiene fichero de nombre fileName.
	 * @param i Posicin del fichero a obtener.
	 * @return GeoRasterFile.
	 */
	public GeoRasterFile getFile(String fileName){
		for(int i=0;i<files.size();i++){
			if(((GeoRasterFile)files.get(i)).getName().equals(fileName))
				return (GeoRasterFile)files.get(i); 
		}
		return null;		
	}
	
	/**
	 * Asigna el nombre al GeoRasterMultiFile
	 * @param name Nombre del GeoRasterMultiFile
	 */
	public void setName(String name){
		this.name = name;
	}
	
	/**
	 * Obtiene la lista de bandas
	 * @return BandList
	 */
	public BandList getBands(){
		return bandList;
	}
	
	/**
	 * Obtiene la coordenada X mnima de toda la lista
	 * @return Coordenada X mnima
	 */
	public double getMinX(){
		double minx = Double.MAX_VALUE;
		for(int i = 0; i < files.size(); i++){
			double aux = ((GeoRasterFile)files.get(i)).getExtent().getMin().getX();
			if(aux < minx)
				minx = aux;
		}
		return minx;
	}
	
	/**
	 * Obtiene la coordenada Y mnima de toda la lista
	 * @return Coordenada Y mnima
	 */
	public double getMinY(){
		double miny = Double.MAX_VALUE;
		for(int i = 0; i < files.size(); i++){
			double aux = ((GeoRasterFile)files.get(i)).getExtent().getMin().getY();
			if(aux < miny)
				miny = aux;
		}
		return miny;
	}
	
	/**
	 * Obtiene la coordenada Y mxima de toda la lista
	 * @return Coordenada Y mxima
	 */
	public double getMaxX(){
		double maxx = Double.MIN_VALUE;
		for(int i = 0; i < files.size(); i++){
			double aux = ((GeoRasterFile)files.get(i)).getExtent().getMin().getY();
			if(aux > maxx)
				maxx = aux;
		}
		return maxx;
	}

	/**
	 * Obtiene la coordenada Y mxima de toda la lista
	 * @return Coordenada Y mxima
	 */
	public double getMaxY(){
		double maxy = Double.MIN_VALUE;
		for(int i = 0; i < files.size(); i++){
			double aux = ((GeoRasterFile)files.get(i)).getExtent().getMin().getY();
			if(aux > maxy)
				maxy = aux;
		}
		return maxy;
	}
	
	/**
	 * Obtiene el extent del multi fichero. Este corresponde al primer
	 * GeoRasterFile de la lista.
	 * @return Extent
	 */
	public Extent getExtent(){
		if(files.size() == 0)
			return null;
		else
			return ((GeoRasterFile)files.get(0)).getExtent();
	}
	
	/**
	 * Convierte un punto desde coordenadas pixel a coordenadas del mundo.
	 * @param pt Punto a transformar
	 * @return punto transformado en coordenadas del mundo
	 */
	public Point2D rasterToWorld(Point2D pt) {
		return ((GeoRasterFile)files.get(0)).rasterToWorld(pt);
	}
	
	/**
	 * Convierte un punto desde del mundo a coordenadas pixel.
	 * @param pt Punto a transformar
	 * @return punto transformado en coordenadas pixel
	 */
	public Point2D worldToRaster(Point2D pt) {
		return ((GeoRasterFile)files.get(0)).worldToRaster(pt);
	}
}
