/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2006 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 */
package org.cresques.io.data;

import java.awt.geom.Point2D;

import org.cresques.io.GeoRasterFile;
import org.cresques.io.datastruct.Histogram;
import org.cresques.io.datastruct.Statistic;
import org.cresques.io.exceptions.FileFoundInListException;
import org.cresques.px.Extent;
import org.cresques.px.PxRaster;
import org.cresques.util.Utilities;
import org.cresques.filter.IRaster;

public class Grid implements IQueryableRaster, IWritableRaster  {
	
	private GeoRasterMultiFile 		grmf = new GeoRasterMultiFile(null);
	private RasterBuf				rasterBuf = null;
	private Histogram				histogram = null;
	private PxRaster				pxRaster = null;
	private Statistic				stats = null;
	protected int					width = 0;
	protected int					height = 0;
	protected Extent				dataExtent = null;
	
	private double m_dNoDataValue = -99999; //TODO:temporal
	
	/**
	 * Constructor
	 */
	public Grid(){}
	
	/**
	 * Constructor
	 * @param grf Lista de geoRasterFile
	 */
	public Grid(GeoRasterFile[] grf){
		for(int i = 0; i< grf.length; i++)
			addFile(grf[i]);
	}
	
	/**
	 * Constructor
	 * @param grf GeoRasterFile
	 */
	public Grid(GeoRasterFile grf){
		addFile(grf);
	}

	/**
	 * Asigna el renderizador para el calculo de histogramas a partir 
	 * de la visualizacin.
	 * @param pxR Renderizador.
	 */
	public void addRenderizer(PxRaster pxR){
		this.pxRaster = pxR;
		this.stats = pxR.stats;
		this.getHistogram().addRenderizer(pxR);
	}
	
	/**
	 * Aade un GeoRasterFile al Grid
	 * @param grf GeoRasterFile a aadir
	 */
	public void addFile(GeoRasterFile grf){
		try{
			grmf.addFile(grf);
			width = grf.getWidth();
			height = grf.getHeight();
		}catch(FileFoundInListException e){
			//El fichero est en la lista por lo que no lo aadimos
		}
	}
	
	/**
	 * Elimina un GeoRasterFile del Grid
	 * @param grf GeoRasterFile a eliminar
	 */
	public void removeFile(GeoRasterFile grf){
		grmf.removeFile(grf);
	}
	
	/**
	 * Obtiene la estructura que contiene la lista de ficheros del Grid
	 * @return GeoRasterMultiFile
	 */
	public GeoRasterMultiFile getGeoRasterMultiFile(){
		return grmf;
	}
	
	/**
	 * Libera el buffer de memoria
	 */
	public void free(){
		rasterBuf = null;
		System.gc();
	}
	
	/**
	 * Resetea la asignacin de dibujado de las bandas de la imagen
	 * sobre el DataImage cuando se hace un update para esta banda.
	 */
	public void clearDrawableBand(){
		for(int i = 0; i< grmf.getFileCount(); i++)
			grmf.getBands().clearDrawableBand();
	}
	
	/**
	 * Para este GeoRasterFile asigna que bandas se pintaran
	 * sobre el RasterBuf cuando se haga un update. Cada posicin del vector es una banda
	 * del rasterBuf y el contenido de esa posicin es la banda de la imagen que se dibujar
	 * sobre ese RasterBuf.
	 * @param drawableBands	.
	 */
	public void addDrawableBands(int[] drawableBands){
		for(int i = 0; i< drawableBands.length; i++)
			grmf.getBands().addDrawableBand(i, drawableBands[i]);				
	}
	
	/**
	 * Asigna el rea de interes en coordenadas del mundo real. Si las coordenadas exceden del tamao de la imagen
	 * estas coordenadas son ajustadas el extent.
	 * @param x Coordenada X, esquina superior izquierda
	 * @param y Coordenada Y, esquina superior izquierda
	 * @param w Ancho del rea 
	 * @param h Alto del rea
	 * @throws ArrayIndexOutOfBoundsException
	 */
	public void setAreaOfInterest(double x, double y, double w, double h) throws ArrayIndexOutOfBoundsException{
		dataExtent = new Extent(x, y, x + w, y - h);
		Extent fitExtent = Utilities.calculateAdjustedView(dataExtent, grmf.getExtent());
		
		rasterBuf = grmf.getWindowRaster(fitExtent.minX(), fitExtent.maxY(), fitExtent.width(), fitExtent.height());
		
		if(!Utilities.isInside(dataExtent, fitExtent)){
			double distWcX = Math.abs(fitExtent.getMin().getX() - dataExtent.getMin().getX());
			double distWcY = Math.abs(fitExtent.getMax().getY() - dataExtent.getMax().getY());
			double mtsPixelX = grmf.getExtent().width() / grmf.getWidth();
			double mtsPixelY = grmf.getExtent().width() / grmf.getWidth(); 
			double pxX = distWcX / mtsPixelX;
			double pxY =  distWcY / mtsPixelY;
			
			double wPx = dataExtent.width() / mtsPixelX;
			double hPx = dataExtent.height() / mtsPixelY;
			
			RasterBuf buf = new RasterBuf(rasterBuf.getDataType(), (int)wPx, (int)hPx, rasterBuf.getBandCount(), true);
			buf.setToNoData();
			
			for (int iBand = 0; iBand < rasterBuf.getBandCount(); iBand++)	
				for (int row = 0; row < rasterBuf.getHeight(); row++) 
					for (int col = 0; col < rasterBuf.getWidth(); col++) 
						buf.setElemByte((int)(row + pxY), (int)(col + pxX), 
										iBand, 
										rasterBuf.getElemByte(row, col, iBand));
			rasterBuf = buf;
		}
	}
	
	/**
	 * Asigna el rea de interes en coordenadas pixel. Si las coordenadas exceden del tamao de la imagen
	 * lanza una excepcin.
	 * @param x Coordenada X, esquina superior izquierda
	 * @param y Coordenada Y, esquina superior izquierda
	 * @param w Ancho del rea 
	 * @param h Alto del rea
	 * @throws ArrayIndexOutOfBoundsException
	 */
	public void setAreaOfInterest(int x, int y, int w, int h) throws ArrayIndexOutOfBoundsException{
		if((x + w) > getWidth() || (y + h) > getHeight())
			throw new ArrayIndexOutOfBoundsException("Out of Image");
		
		Point2D begin = grmf.rasterToWorld(new Point2D.Double(x, y));
		Point2D end = grmf.rasterToWorld(new Point2D.Double(x + w, y + h));
		if(begin != null && end != null)
			dataExtent = new Extent(begin, end);
		
		rasterBuf = grmf.getWindowRaster(x, y, w, h);
		//width = rasterBuf.getWidth();
		//height = rasterBuf.getHeight();
	}
	
	/*
	 *  (non-Javadoc)
	 * @see org.gvsig.fmap.dataaccess.IQueryableRaster#setAreaOfInterest(double, double, double, double, int, int)
	 */
	public int[] setAreaOfInterest(double minX, double minY, double maxX, double maxY, int bufWidth, int bufHeight) throws ArrayIndexOutOfBoundsException{
		//TODO: VALIDACIN: Comprobacin de exceso de memoria reservada
		dataExtent = new Extent(minX, minY, maxX, maxY);
		Extent fitExtent = Utilities.calculateAdjustedView(dataExtent, grmf.getExtent());
		
		//Caso en que la extensin pedida SI se sale fuera del rea del raster
		//El pixel a medias entre NoData y datos se pone NoData
		if(Utilities.isOutside(dataExtent, grmf.getExtent())) {
			double error = 0.01;
			//Upper Left
			double distWcX = Math.abs(fitExtent.getMin().getX() - dataExtent.getMin().getX());
			distWcX = (distWcX > error) ? distWcX : 0;
			double distWcY = Math.abs(fitExtent.getMax().getY() - dataExtent.getMax().getY());
			distWcY = (distWcY > error) ? distWcY : 0;
			//Pixel inicial del buffer donde se empieza a dibujar. Redondeamos por arriba pq lo que sobra se pone NoData
			double initPxX = Math.ceil((distWcX * bufWidth) / (Math.abs(maxX - minX))); 
			double initPxY = Math.ceil((distWcY * bufHeight) / (Math.abs(maxY - minY)));
			//Obtenemos la coordenada real para el pixel inicial
			double wcXInit = (minX < grmf.getExtent().minX()) ? grmf.getExtent().minX() : minX;
			double wcYInit = (maxY > grmf.getExtent().maxY()) ? grmf.getExtent().maxY() : maxY;
			
			//Lower Right
			distWcX = Math.abs(fitExtent.getMax().getX() - dataExtent.getMin().getX());
			distWcX = (distWcX > error) ? distWcX : 0;
			distWcY = Math.abs(fitExtent.getMin().getY() - dataExtent.getMax().getY());
			distWcY = (distWcY > error) ? distWcY : 0;
			//Pixel final del buffer donde se dibuja. Redondeamos por abajo pq lo que sobra se pone NoData
			double endPxX = Math.floor((distWcX * bufWidth) / (Math.abs(maxX - minX))); 
			double endPxY = Math.floor((distWcY * bufHeight) / (Math.abs(maxY - minY)));
			//Obtenemos la coordenada real para el pixel
			double wcXEnd = (maxX > grmf.getExtent().maxX()) ? grmf.getExtent().maxX() : maxX;
			double wcYEnd = (minY < grmf.getExtent().minY()) ? grmf.getExtent().minY() : minY;
			
			int copyX = (int)Math.abs(endPxX - initPxX);
			int copyY = (int)Math.abs(endPxY - initPxY);
			
			rasterBuf = grmf.getWindowRaster(wcXInit, wcYEnd, wcXEnd, wcYInit, copyX, copyY);
			RasterBuf buf = new RasterBuf(rasterBuf.getDataType(), bufWidth, bufHeight, rasterBuf.getBandCount(), true);
			buf.setNoDataValue(m_dNoDataValue);
			buf.setToNoData();
			
			switch(rasterBuf.getDataType()) {
			  case IRaster.TYPE_BYTE:
				  for (int iBand = 0; iBand < rasterBuf.getBandCount(); iBand++)	
						for (int row = 0; row < copyY; row++) 
							for (int col = 0; col < copyX; col++) 
								buf.setElemByte((int)(row + initPxY), (int)(col + initPxX), 
												iBand, 
												rasterBuf.getElemByte(row, col, iBand));
				    break;
			  case IRaster.TYPE_SHORT:
			  case IRaster.TYPE_USHORT:
				  for (int iBand = 0; iBand < rasterBuf.getBandCount(); iBand++)	
						for (int row = 0; row < copyY; row++) 
							for (int col = 0; col < copyX; col++) 
								buf.setElemShort((int)(row + initPxY), (int)(col + initPxX), 
												iBand, 
												rasterBuf.getElemShort(row, col, iBand));
				    break;
			  case IRaster.TYPE_INT:
				  for (int iBand = 0; iBand < rasterBuf.getBandCount(); iBand++)	
						for (int row = 0; row < copyY; row++) 
							for (int col = 0; col < copyX; col++) 
								buf.setElemInt((int)(row + initPxY), (int)(col + initPxX), 
												iBand, 
												rasterBuf.getElemInt(row, col, iBand));
				    break;
			  case IRaster.TYPE_FLOAT:
				  for (int iBand = 0; iBand < rasterBuf.getBandCount(); iBand++)	
						for (int row = 0; row < copyY; row++) 
							for (int col = 0; col < copyX; col++) 
								buf.setElemFloat((int)(row + initPxY), (int)(col + initPxX), 
												iBand, 
												rasterBuf.getElemFloat(row, col, iBand));
				    break;
			  case IRaster.TYPE_DOUBLE:
				for (int iBand = 0; iBand < rasterBuf.getBandCount(); iBand++)	
					for (int row = 0; row < copyY; row++) 
						for (int col = 0; col < copyX; col++) 
							buf.setElemDouble((int)(row + initPxY), (int)(col + initPxX), 
											iBand, 
											rasterBuf.getElemDouble(row, col, iBand));
			    break;
			}
			rasterBuf = buf;
			return null;
			
		}
		
		//Caso en que la extensin pedida NO se sale fuera del rea del raster
		rasterBuf = grmf.getWindowRaster(fitExtent.minX(), fitExtent.minY(), fitExtent.maxX(), fitExtent.maxY(), bufWidth, bufHeight);
		return null;
	}
	
	/*
	 *  (non-Javadoc)
	 * @see org.gvsig.fmap.dataaccess.IQueryableRaster#setAreaOfInterest(double, double, double, double)
	 */
	public void setAreaOfInterestWithNoData(double x, double y, double w, double h) throws ArrayIndexOutOfBoundsException{
		dataExtent = new Extent(x, y, x + w, y - h);
		rasterBuf = grmf.getWindowRasterWithNoData(dataExtent.getMin().getX(), dataExtent.getMax().getY(), w, h);
	}
		
	/**
	 * Ajusta el rea del grid a un ancho y un alto dado en pixeles. Este ajuste se har 
	 * en relacin a un mtodo de interpolacin definido en el parmetro.
	 * @param w Ancho de la nueva imagen.
	 * @param h Alto de la nueva imagen.
	 * @param interpolation Mtodo de interpolacin que se usar en el ajuste.
	 * @param bands Bandas que se desea en el nuevo RasterBuf ajustado. Si este parmetro es null
	 * se usarn todas las bandas de la imagen.
	 */
	public void setAdjustedWindow(int w, int h, int interpolationMethod, int[] bands){
		if(w == width && h == height)
			return;
		switch(interpolationMethod){
		case IQueryableRaster.INTERPOLATION_NearestNeighbour:
				rasterBuf = rasterBuf.adjustRasterNearestNeighbour(w, h, bands);
				break;
		case IQueryableRaster.INTERPOLATION_Bilinear:
				System.out.println("Method not implemented yet");
				break;
		case IQueryableRaster.INTERPOLATION_InverseDistance:
				System.out.println("Method not implemented yet");
				break;
		case IQueryableRaster.INTERPOLATION_BicubicSpline:
				System.out.println("Method not implemented yet");
				break;
		case IQueryableRaster.INTERPOLATION_BSpline:
				System.out.println("Method not implemented yet");
				break;
		}
		//width = rasterBuf.getWidth();
		//height = rasterBuf.getHeight();
	}
	
	public RasterBuf getData(int x, int y, int w, int h) {

		return null;
	}

	/**
	 * Obtiene el tipo de datos del grid.
	 * @return entero que representa el tipo de dato de las celdas del grid.
	 */
	public int getDataType() {
		return grmf.getDataType()[0];
	}

	/**
	 * Obtiene la altura del grid.
	 * @return altura en celdas del grid.
	 */
	public int getHeight() {
		return height;
	}

	/**
	 * Obtiene la anchura del grid.
	 * @return anchura en celdas del grid.
	 */
	public int getWidth() {
		return width;
	}

	public double getXCellSize() {

		//TODO:remove this
		Extent e = grmf.getExtent();
		double dCellsize = (e.getMax().getX() - e.getMin().getX() ) / (double) width;
		
		return dCellsize;
		
	}

	public double getYCellSize() {

		return 0;
	}

	public boolean isNoData(int x, int y, int band) {

		return false;
		
	}
	
	public void setNoDataValue(double dNoDataValue){
		
		//TODO:temporal
		m_dNoDataValue = dNoDataValue;
	}
	
	public double getNoDataValue(){
		
		return m_dNoDataValue;
		
	}

	public void setDataType(int dt) {
		
	}
	
	public int getBandCount(){
		if(rasterBuf != null)
			return rasterBuf.getBandCount();
		else
			return this.getGeoRasterMultiFile().getBandCount();
	}

	/**
	 * Obtiene el buffer de datos del grid
	 * @return RasterBuf
	 */
	public RasterBuf getRasterBuf() {
		return rasterBuf;
	}
	
	/**
	 * Obtiene el objeto histograma asociado al grid
	 * @return
	 */
	public Histogram getHistogram() {
		if(histogram == null)
			histogram = new Histogram(this);
		
		return histogram;
	}

	/**
	 * Obtiene la clase para estadisticas
	 * @return
	 */
	public Statistic getStatistic() {
		return stats;
	}
	
	/**
	 * Evalua si una coordenada pixel cae dentro de la imagen.
	 * @param x coordenada pixel X
	 * @param y coordenada pixel Y
	 */
	public boolean isPixelInGrid(int x, int y){
		return (x >= 0 && y >= 0 && x <= getWidth() && y <= getHeight());
	}
	
	/**
	 * Extent de la fuente de datos
	 */
	public Extent getExtent(){
		return grmf.getExtent();
	}
	
	/**
	 * Extent del grid
	 */
	public Extent getDataExtent(){
		return dataExtent;
	}
}

