
 
package org.agil.core.spatialindex;

import com.vividsolutions.jts.geom.Envelope;

import java.util.*;
import java.io.IOException;
/**
 * Ttulo: FileRtreeSpatialIndex
 * Descripcion:
 *
 * <p>
 * Implementacin del indice espacial RTree que funciona
 * con un fichero de indice.
 * <br>
 * Esta basado en la libreria RTree de Marios Hadjieleftheriou
 * (por si hay que buscar en el Google)<br>
 * </p>
 * <br>
 * <p>
 * Como funciona el indice?<br>
 *  Indexa posiciones de objetos, las del almacenamiento real,
 * en funcion de rectangulos asociados a estos objetos.<br>
 * Por tanto, <b>no esperemos que nos devuelva los objetos que guardamos</b>, sino
 * las posiciones donde estos objetos estan guardados.<br>
 * La clase que haga uso de este indice, una vez obtenido una lista con las posiciones
 * de los objetos, debe recuperarlos de su almacenamiento real.<br>
 * As, por ejemplo, si los objetos estan guardados en un fichero Shapefile,
 * con el indice espacial obtenemos la posicion del registro en el fichero shx.<br>
 * Accediendo a esta posicion, obtenemos la posicion en el fichero geometrico (shp)
 * y en el alfanumerico (dbf).<br>
 * En el caso de bases de datos, la posicion devuelta es directamente el GID, o
 * el ROWID de Oracle.
 * </p>
 *
 *
 
 */

public class FileRtreeSpatialIndex implements FileSpatialIndex {

  /**
   * Directorio donde se encuentra el fichero de indice.
   * Lo ideal es que se agrupe a nivel de dataAccesors, es decir, si una FeatureCollection
   * toma sus datos de Oracle, debe haber un directorio oradatasources/ donde guardarse
   * estos indices. As, si hay varios origenes de datos (shapefiledatasources/) sus
   * indices espaciales no se machacan entre si.
   * */
  private String pathSpatialIndex;
  /**
   * Nombre del fichero
   */
  private String featureCollectionName;

  /**
   * Clase que verdaderamente hace el trabajo de indexado espacial.
   * */
  private RTree fileRTree;

  /**
   * Constructor: busca en la ruta especificada y se prepara para leer y escribir
   * en el fichero de indice. El unico requisito para poder trabajar es que
   * el fichero de indice se haya creado alguna vez (CREATE SPATIAL INDEX en SQL)
   * con el mtodo createSpatialIndex.
   *
   * @param pathSpatialIndex directorio donde se ubica el indice espacial
   * @param featureCollectionName nombre de la featureCollection al que pertenece el indice
   *
   * @see FileRtreeSpatialIndex.createSpatialIndex
   * */
  public FileRtreeSpatialIndex(String pathSpatialIndex,String featureCollectionName) throws IOException {
    this.pathSpatialIndex = pathSpatialIndex;
    this.featureCollectionName = featureCollectionName;
    fileRTree = new RTree(pathSpatialIndex,featureCollectionName);
    fileRTree.open();
  }

  /**
   * Metodo que se encarga de crear, solo la primera vez, el fichero de indice espacial.
   * Lo normal es que esto se haga desde scripts bath (en un proceso de carga de indices)
   * o por peticion.
   * */
  public void createSpatialIndex() throws IOException{
    if (fileRTree != null){
       fileRTree.create();
    }
  }

  /**
   * Abre el fichero de indice espacial
   * */
  public void open() throws IOException{
    if (fileRTree != null)
      fileRTree.open();
  }

  /**
   * Libera los recursos consumidos -canales de I/O- por el fichero
   * de indice
   * */
  public void close() throws IOException{
   if (fileRTree != null)
			  fileRTree.close();
  }

  /**
   * Indexa la posicion (representada por Object, que debe ser un Long)
   * donde se guarda el objeto  que se quiere indexar.
   *
   * @param itemEnv rectangulo por el que se indexa la posicion del objeto
   * @param item Long que nos dice donde debemos buscar el objeto en el origen
   * de datos principal (no olvidemos que esto solo es un indice)
   *
   * @see com.vividsolutions.jts.index.SpatialIndex
   * */
  public void insert(Envelope itemEnv, Object item) throws IOException {
    Long posicion = (Long) item;
    fileRTree.write(itemEnv,posicion.longValue());
  }

  /**
   * Devuelve la posicion -asociada a un fichero de almacenamiento principal-
   * de los objetos cuyos Envelopes solapen con el especificado
   *
   * Problema: EL DEVOLVER UN LIST PUEDE DEJAR SIN MEMORIA. Quieras o no,
   * si nuestra consulta al indice devuelve millones de candidatos, meterlos todos en
   * memoria consumen recursos. (Un Long es un objeto, y como tal consume recursos)
   *
   * @param searchEnv rectangulo de consulta del que se quieren obtener todos los objetos
   *  que solapen con el
   *
   * @return List que contiene tantos Long como objetos solapen con el rectangulo
   * @throws excepcion de entrada/salida por problemas en el acceso al fichero de indice
   *
   * */
  public List query(Envelope searchEnv) throws IOException {
    ArrayList solucion = new ArrayList();
    Iterator iterator = fileRTree.iterator(searchEnv);
    while(iterator.hasNext()){
      Long posicion = (Long)iterator.next();
      solucion.add(posicion);
    }//while
    return solucion;
  }
  /**
   * Similar al anterior pero devuelve un iterador sobre los candidatos.
   * (Esto consume menos recursos)
   * @param searchEnv
   * @return
   * @throws IOException
   */
  public Iterator queryIterator(Envelope searchEnv) throws IOException {
    return fileRTree.iterator(searchEnv);
  }
  /**
   * Aadimos un metodo para borrar items del indice espacial
   * @param envelope
   * @param item
   */
  public void delete(Envelope envelope, Object item){
    fileRTree.delete(envelope, ((Long)item).longValue());
  }

  public void update(Envelope envelope, Object oldPos, Object newPos){
    Long oldPosL = (Long)oldPos;
    Long newPosL = (Long)newPos;
    fileRTree.update(envelope, oldPosL.longValue(), newPosL.longValue());

  }


}