package org.agil.core.spatialindex;

/**
 *  Nodo interno del RTree.
 *
 *@author    Chris Seguin
 */
public class Index extends AbstractNode {
	/**
	 *  Construye un nuevo Index.
	 *
	 *@param  tree        instancia de RTree al que pertenece este Index
	 *@param  parent      padre del que depende
	 *@param  pageNumber  pagina donde deberia almacenarse
	 *@param  level       nivel donde deberia ser insertado
	 */
	protected Index(RTree tree, long parent, long pageNumber, int level) {
		super(tree, parent, pageNumber, level);
	}


	/**
	 *  Retrieves the <B>i-th</B> child node. Loads one page into main memory.
	 *
	 *@param  i  The index of the child in the data array.
	 *@return    The i-th child.
	 */
	public AbstractNode getChild(int i) {
		/*
		 *  if (i < 0 || i >= usedSpace) {
		 *  throw new IndexOutOfBoundsException("" + i);
		 *  }
		 */
		return tree.file.readNode(branches[i]);
	}


	/**
	 *  Retorna aquella hoja que este contenida en el HyperCube.
	 *
	 *@param  h  HyperCube de consulta
	 *@return    hoja encontrada
	 */
	protected Leaf chooseLeaf(HyperCube h) {
		int i;
		switch (tree.getTreeType()) {
			case RTree.RTREE_LINEAR:
			case RTree.RTREE_QUADRATIC:
				i = findLeastEnlargement(h);
				break;
			case RTree.RSTAR:
				if (level == 1) {
					// if this node points to leaves...
					i = findLeastOverlap(h);
				}
				else {
					i = findLeastEnlargement(h);
				}
				break;
			default:
				throw new IllegalStateException("Invalid tree type.");
		}
		return ((AbstractNode) getChild(i)).chooseLeaf(h);
	}


	/**
	 *  Busca una hoja que contenga el HyperCube <b>h</b> .
	 *
	 *@param  h  HyperCube de consulta
	 *@return    hoja encontrada
	 */
	protected Leaf findLeaf(HyperCube h) {
		for (int i = 0; i < usedSpace; i++) {
			if (data[i].contains(h)) {
				Leaf l = ((AbstractNode) getChild(i)).findLeaf(h);
				if (l != null) {
					return l;
				}
			}
		}
		return null;
	}


	/**
	 *  Busca una hoja que contenga el HyperCube <b>h</b> .
	 *
	 *@param  h    HyperCube de consulta
	 *@param  pos  Description of Parameter
	 *@return      hoja encontrada
	 */
	protected Leaf findLeaf(HyperCube h, long pos) {
		for (int i = 0; i < usedSpace; i++) {
			if (data[i].contains(h)) {
				Leaf l = ((AbstractNode) getChild(i)).findLeaf(h, pos);
				if (l != null) {
					return l;
				}
			}
		}
		return null;
	}


	/**
	 *  Inserts a new node into the tree. If enough space is available the
	 *  insertion is straightforward. If not, a split must occur.
	 *
	 *@param  node  The node that should be inserted.
	 *@return       True if a split occurred, false otherwise.
	 */
	protected boolean insert(AbstractNode node) {
		if (usedSpace < tree.getNodeCapacity()) {
			addData(node);
			node.parent = pageNumber;
			tree.file.writeNode(node);
			tree.file.writeNode(this);
			Index p = (Index) getParent();
			if (p != null) {
				p.adjustTree(this, null);
			}
			return false;
		}
		else {
			Index[] a = splitIndex(node);
			Index n = a[0];
			Index nn = a[1];
			if (isRoot()) {
				n.parent = 0;
				n.pageNumber = -1;
				nn.parent = 0;
				nn.pageNumber = -1;
				long p = tree.file.writeNode(n);
				for (int i = 0; i < n.usedSpace; i++) {
					AbstractNode ch = (AbstractNode) n.getChild(i);
					ch.parent = p;
					tree.file.writeNode(ch);
				}
				p = tree.file.writeNode(nn);
				for (int i = 0; i < nn.usedSpace; i++) {
					AbstractNode ch = (AbstractNode) nn.getChild(i);
					ch.parent = p;
					tree.file.writeNode(ch);
				}
				Index r = new Index(tree, RTree.NIL, 0, level + 1);
				r.addData(n.nodeMbb, n.pageNumber);
				r.addData(nn.nodeMbb, nn.pageNumber);
				tree.file.writeNode(r);
			}
			else {
				n.pageNumber = pageNumber;
				n.parent = parent;
				nn.pageNumber = -1;
				nn.parent = parent;
				tree.file.writeNode(n);
				long p = tree.file.writeNode(nn);
				for (int i = 0; i < nn.usedSpace; i++) {
					AbstractNode ch = (AbstractNode) nn.getChild(i);
					ch.parent = p;
					tree.file.writeNode(ch);
				}
				Index pr = (Index) getParent();
				pr.adjustTree(n, nn);
			}
			tree.stats.splits++;
			return true;
		}
	}


	/**
	 *  Called by insert to adjust the parents of the node that was modified by
	 *  the insertion. Stops when it reaches the root.
	 *
	 *@param  n1  The child node that caused the need for adjustment.
	 *@param  n2  If a split occured in the child node, this is the new node that
	 *      must be inserted. Otherwise, this is null.
	 */
	protected void adjustTree(AbstractNode n1, AbstractNode n2) {
		// find entry pointing to old node;
		for (int i = 0; i < usedSpace; i++) {
			if (branches[i] == n1.pageNumber) {
				data[i] = n1.getNodeMbb();
				break;
			}
		}
		updateNodeMbb();
		tree.file.writeNode(this);
		// if a split has occured we must insert the new node, else we must continue adjusting the tree
		// until we hit the root.
		if (n2 != null) {
			insert(n2);
		}
		else if (!isRoot()) {
			Index p = (Index) getParent();
			p.adjustTree(this, null);
		}
	}


	/**
	 *  Anade un nuevo HyperCube a todos los MBBs presentes en este nodo. Calcula
	 *  la diferencia de areas y escoge aquella entrada que lo contenga con la
	 *  menor diferencia. Basandonos en esta medida, escogemos el camino que
	 *  sostendra el nuevo HyperCube.
	 *
	 *@param  h  Description of Parameter
	 *@return    Description of the Returned Value
	 */
	private int findLeastEnlargement(HyperCube h) {
		double area = Double.POSITIVE_INFINITY;
		int sel = -1;
		for (int i = 0; i < usedSpace; i++) {
			double enl = data[i].combinedMbb(h).getArea() - data[i].getArea();
			if (enl < area) {
				area = enl;
				sel = i;
			}
			else if (enl == area) {
				sel = (data[sel].getArea() <= data[i].getArea()) ? sel : i;
			}
		}
		return sel;
	}


	/**
	 *  R*-tree criterion for choosing the best branch to follow. [Beckmann,
	 *  Kriegel, Schneider, Seeger 'The R*-tree: An efficient and Robust Access
	 *  Method for Points and Rectangles]
	 *
	 */
	/**
	 *  Criterio del R*-tree para escoger la mejor rama a seguir. [Beckmann,
	 *  Kriegel, Schneider, Seeger 'The R*-tree: An efficient and Robust Access
	 *  Method for Points and Rectangles]
	 *
	 *@param  h  Description of Parameter
	 *@return    El indice de la rama que apunta a la hoja donde deberia
	 *      insertarse el HyperCube
	 */
	private int findLeastOverlap(HyperCube h) {
		float enlargement = Float.POSITIVE_INFINITY;
		float oo = 0;
		int sel = -1;
		for (int j = 0; j < usedSpace - 1; j++) {
			for (int k = j + 1; k < usedSpace; k++) {
				oo += data[j].intersectingArea(data[k]);
			}
		}
		for (int i = 0; i < usedSpace; i++) {
			float o = 0;
			HyperCube n = data[i].combinedMbb(h);
			for (int j = 0; j < usedSpace; j++) {
				if (j == i) {
					continue;
				}
				o += data[j].intersectingArea(n);
			}
			for (int j = 0; j < usedSpace - 1; j++) {
				if (j == i) {
					continue;
				}
				for (int k = j + 1; k < usedSpace; k++) {
					if (k == i) {
						continue;
					}
					o += data[j].intersectingArea(data[k]);
				}
			}
			float enl = o - oo;
			if (enl < enlargement) {
				enlargement = enl;
				sel = i;
			}
			else if (enl == enlargement) {
				double area1 = n.getArea() - data[i].getArea();
				double area2 = data[sel].combinedMbb(h).getArea() - data[sel].getArea();
				if (area1 == area2) {
					sel = (data[i].getArea() <= data[sel].getArea()) ? i : sel;
				}
				else {
					sel = (area1 < area2) ? i : sel;
				}
			}
		}
		return sel;
	}


	/**
	 *  Description of the Method
	 *
	 *@param  n  Description of Parameter
	 *@return    Description of the Returned Value
	 */
	private Index[] splitIndex(AbstractNode n) {
		int[][] group = null;
		switch (tree.getTreeType()) {
			case RTree.RTREE_LINEAR:
			case RTree.RTREE_QUADRATIC:
				group = rtreeSplit(n.getNodeMbb(), n.pageNumber);
				break;
			case RTree.RSTAR:
				break;
			default:
				throw new IllegalStateException("Invalid tree type.");
		}
		Index i1 = new Index(tree, parent, pageNumber, level);
		Index i2 = new Index(tree, parent, -1, level);
		int[] g1 = group[0];
		int[] g2 = group[1];
		for (int i = 0; i < g1.length; i++) {
			i1.addData(data[g1[i]], branches[g1[i]]);
		}
		for (int i = 0; i < g2.length; i++) {
			i2.addData(data[g2[i]], branches[g2[i]]);
		}
		return new Index[]{i1, i2};
	}
}
// Index
